﻿Imports System.IO
Structure UndoData
    Public Sub New(ByVal i As Integer, ByVal c As Color)
        Index = i
        ColorCode = c
    End Sub
    Dim Index As Integer
    Dim ColorCode As Color
End Structure

Class MainWindow

    Dim SelectColor As Color    ' 選択された色
    Dim SourceArray() As Byte   ' 塗り絵の元になるイメージ
    Dim DestArray() As Byte
    Dim SourceStride As Integer
    Dim SourceWidth As Double
    Dim SourceHeight As Double
    Dim MapColors() As Color
    Dim UndoBuffer As List(Of UndoData)

    Private Sub Button1_Click(ByVal sender As System.Object, ByVal e As System.Windows.RoutedEventArgs) Handles Button1.Click

        InitPalette()

    End Sub

    ' 色の対応テーブルを初期化
    Sub InitPalette()

        MapColors = New Color(15) {}
        For i = 0 To 15
            MapColors(i) = If(i = 0, Colors.Black, Colors.White)
        Next
        DisplayNurie()
        UndoBuffer.Clear()

    End Sub

    ' クレヨンの色を取得
    Function SelectCrayon(ByVal c As Integer) As Color

        Dim CrayonColors As Color() = New Color() {Color.FromRgb(255, 0, 0), Color.FromRgb(255, 255, 255), Color.FromRgb(255, 192, 128), Color.FromRgb(192, 192, 192),
                                                   Color.FromRgb(156, 156, 48), Color.FromRgb(20, 20, 20), Color.FromRgb(163, 112, 65), Color.FromRgb(255, 128, 128),
                                                   Color.FromRgb(120, 60, 48), Color.FromRgb(120, 120, 34), Color.FromRgb(244, 150, 60), Color.FromRgb(192, 128, 255),
                                                   Color.FromRgb(0, 255, 0), Color.FromRgb(0, 0, 255), Color.FromRgb(128, 255, 255), Color.FromRgb(255, 255, 128),
                                                   Color.FromRgb(90, 156, 156)}

        Return CrayonColors(c)

    End Function

    ' ファイルを読み込んでSourceArrayに設定
    Sub LoadBitmap(ByVal FileName As String)

        Dim s As Stream = Application.GetResourceStream(New Uri("Images/" & FileName, UriKind.Relative)).Stream
        Dim bit As BitmapImage = New BitmapImage()
        bit.BeginInit()
        bit.StreamSource = s
        bit.EndInit()

        SourceWidth = bit.PixelWidth
        SourceHeight = bit.PixelHeight

        SourceStride = (SourceWidth * bit.Format.BitsPerPixel + 7) \ 8
        SourceArray = New Byte(SourceStride * SourceHeight) {}
        bit.CopyPixels(SourceArray, SourceStride, 0)

    End Sub

    ' データを元に塗り絵イメージを表示
    Sub DisplayNurie()

        Dim wb As WriteableBitmap = New WriteableBitmap(SourceWidth, SourceHeight, 96, 96, PixelFormats.Bgr32, Nothing)
        Dim DestStride As Integer = (SourceWidth * wb.Format.BitsPerPixel + 7) \ 8

        DestArray = New Byte(DestStride * SourceHeight) {}

        Dim i As Integer = 0
        For y As Integer = 0 To SourceHeight - 1
            For x As Integer = 0 To SourceWidth - 1
                Dim g As Integer = GetIndex(SourceArray, x, y, SourceStride)
                i = y * DestStride + x * ((wb.Format.BitsPerPixel + 7) \ 8)
                DSetPixel(DestArray, i, MapColors(g))
            Next
        Next

        Dim r As Int32Rect = New Int32Rect(0, 0, SourceWidth, SourceHeight)
        wb.WritePixels(r, DestArray, DestStride, 0)
        Image1.Source = wb

    End Sub

    ' 元イメージからインデックス番号を取得(0～15)
    Function GetIndex(ByRef a As Byte(), ByVal x As Integer, ByVal y As Integer, ByVal stride As Integer) As Integer

        Dim Index As Integer = y * stride + x \ 2
        If x Mod 2 = 0 Then
            Return a(Index) \ 16
        Else
            Return a(Index) And 15
        End If

    End Function

    ' 所定の位置に色を設定
    Function DSetPixel(ByRef a As Byte(), ByVal i As Integer, ByVal c As Color) As Integer

        a(i) = c.B
        a(i + 1) = c.G
        a(i + 2) = c.R
        a(i + 3) = 255

        Return 0

    End Function

    Private Sub Image1_MouseLeftButtonDown(ByVal sender As System.Object, ByVal e As System.Windows.Input.MouseButtonEventArgs) Handles Image1.MouseLeftButtonDown

        Dim p As Point = e.GetPosition(Image1)
        Dim i As Integer = GetIndex(SourceArray, p.X, p.Y, SourceStride)

        ' クリックした場所が黒の場合は何もしない
        If i <> 0 Then
            UndoBuffer.Add(New UndoData(i, MapColors(i)))
            MapColors(i) = SelectColor
            DisplayNurie()
        End If

    End Sub

    ' カーソル位置にクレヨンを表示しない
    Private Sub Canvas1_MouseLeave(ByVal sender As System.Object, ByVal e As System.Windows.Input.MouseEventArgs) Handles Canvas1.MouseLeave

        If e.OriginalSource IsNot Nothing Then
            ImageCursor.Visibility = Windows.Visibility.Hidden
            Canvas1.Cursor = Nothing
        End If

    End Sub

    ' カーソル位置にクレヨンを表示する
    Private Sub Canvas1_MouseEnter(ByVal sender As System.Object, ByVal e As System.Windows.Input.MouseEventArgs) Handles Canvas1.MouseEnter

        If e.OriginalSource IsNot Nothing Then
            ImageCursor.Visibility = Windows.Visibility.Visible
            Canvas1.Cursor = Cursors.None
        End If

    End Sub

    ' クレヨンを移動する
    Private Sub Canvas1_MouseMove(ByVal sender As System.Object, ByVal e As System.Windows.Input.MouseEventArgs) Handles Canvas1.MouseMove

        Dim p As Point = e.GetPosition(Canvas1)
        Canvas.SetLeft(ImageCursor, p.X - 7)
        Canvas.SetTop(ImageCursor, p.Y)

    End Sub

    ' クレヨンを選択
    Private Sub Canvas1_MouseLeftButtonDown(ByVal sender As System.Object, ByVal e As System.Windows.Input.MouseButtonEventArgs)

        If TypeOf (e.OriginalSource) Is Image Then
            For i As Integer = 1 To 17
                Dim MyImage As Image = StackPanel1.FindName("ImageCrayon" & i)
                If e.OriginalSource Is MyImage Then
                    SelectColor = SelectCrayon(i - 1)
                    MyImage.Visibility = Windows.Visibility.Collapsed
                    ImageCursor.Source = MyImage.Source
                ElseIf MyImage.Visibility <> Windows.Visibility.Visible Then
                    MyImage.Visibility = Windows.Visibility.Visible
                End If
            Next
        End If

    End Sub

    ' 顔のイメージがクリックされた
    Private Sub ImageFace_MouseLeftButtonDown(ByVal sender As System.Object, ByVal e As System.Windows.Input.MouseButtonEventArgs) Handles ImageFace.MouseLeftButtonDown

        Dim s As Integer = e.GetPosition(ImageFace).Y \ 40
        SelectBitmap(s)

    End Sub

    ' ビットマップを選択
    Sub SelectBitmap(ByVal c As Integer)

        Dim FileName As String() = New String() {"Tantonuri.bmp", "Imerunuri.bmp", "Upunnuri.bmp", "Urannuri.bmp", "Numannuri.bmp", "Reranuri.bmp"}
        LoadBitmap(FileName(c))
        InitPalette()

    End Sub

    Private Sub Window_Loaded(ByVal sender As System.Object, ByVal e As System.Windows.RoutedEventArgs) Handles MyBase.Loaded

        SelectColor = SelectCrayon(0)
        UndoBuffer = New List(Of UndoData)
        SelectBitmap(0)

    End Sub

    ' アンドゥ
    Private Sub Button2_Click(ByVal sender As System.Object, ByVal e As System.Windows.RoutedEventArgs) Handles Button2.Click

        Dim c As Integer = UndoBuffer.Count - 1
        If c > -1 Then
            MapColors(UndoBuffer(c).Index) = UndoBuffer(c).ColorCode
            DisplayNurie()
            UndoBuffer.RemoveAt(c)
        End If

    End Sub
End Class
